const path = require('path')
const fs = require('fs-extra')
const merge = require('lodash.merge')
const parseJsonConfig = require('parse-json-config')
const resolveFrom = require('resolve-from')
const logger = require('./utils/logger')
const loadConfig = require('./utils/load-config')
const Hooks = require('./hooks')

class Poi {
  constructor(options, config) {
    options = Object.assign(options, {})
    const cliOptions = options.cliOptions || {}
    delete options.cliOptions
    this.options = Object.assign({}, options, {
      baseDir: path.resolve(options.baseDir || '.')
    })
    this.cliOptions = Object.assign({}, cliOptions, {
      args: cliOptions.args || process.argv.slice(2)
    })
    this.config = config || {}
    this.hooks = new Hooks()
    this.logger = logger
    this.loadConfig = loadConfig

    logger.setOptions({
      debug: this.cliOptions.debug
    })

    this.projectPkg = Object.assign(
      { data: {} },
      loadConfig.loadSync({
        files: ['package.json'],
        cwd: this.options.baseDir
      })
    )

    // Expose poi command in env variable
    // In case you wanna use it in config file or somewhere
    process.env.POI_COMMAND = this.options.command
    // TODO: Should we set the default value of `process.env.NODE_ENV` here
    // based on the value of `command` too?
    /* e.g.
    process.env.NODE_ENV =
      process.env.NODE_ENV || this.options.command === 'build'
        ? 'production'
        : this.options.command === 'test'
          ? 'test'
          : 'development'
    */

    if (this.options.configFile !== false) {
      const res = loadConfig.loadSync({
        files:
          typeof this.options.configFile === 'string'
            ? [this.options.configFile]
            : ['poi.config.js', 'package.json'],
        cwd: this.options.baseDir,
        packageKey: 'poi'
      })
      if (res.path) {
        this.configFilePath = res.path
        this.config = merge(res.data, this.config)
        logger.debug(`Poi config file: ${this.configFilePath}`)
      } else {
        logger.debug('Poi is not using any config file')
      }
    }

    this.config = Object.assign(
      {
        // Default values
        outDir: 'dist',
        target: 'app',
        publicPath: '/',
        pluginOptions: {}
      },
      this.config,
      {
        // Proper overrides
        pages: Object.assign({}, this.config.pages, {
          index: Object.assign(
            {
              entry: ['index.js'],
              // Will fallback to `lib/default-template.html` when not found
              template: 'public/index.html',
              chunks: ['chunk-vendors', 'chunk-common', 'index']
            },
            this.config.pages && this.config.pages.index
          )
        }),
        css: Object.assign(
          {
            extract: this.options.command === 'build',
            loaderOptions: {}
          },
          this.config.css
        ),
        devServer: Object.assign(
          {
            host: process.env.HOST || '0.0.0.0',
            port: process.env.PORT || 4000
          },
          this.config.devServer
        ),
        production: Object.assign(
          {
            minimize: true,
            sourceMap: true
          },
          this.config.production
        )
      }
    )

    this.cli = require('cac')()
    this.commands = new Map()
  }

  isCommand(command) {
    if (Array.isArray(command)) {
      return command.includes(this.options.command)
    }
    return command === this.options.command
  }

  createWebpackConfig() {
    const WebpackChain = require('webpack-chain')
    const config = new WebpackChain()

    this.hooks.invoke('chainWebpack', config)

    if (this.cliOptions.inspectWebpack) {
      console.log(config.toString())
      process.exit() // eslint-disable-line unicorn/no-process-exit
    }

    return config.toConfig()
  }

  resolveBaseDir(...args) {
    return path.resolve(this.options.baseDir, ...args)
  }

  chainWebpack(fn) {
    this.hooks.add('chainWebpack', fn)
    return this
  }

  configureDevServer(fn) {
    this.hooks.add('configureDevServer', fn)
    return this
  }

  async bundle() {
    const compiler = require('webpack')(this.createWebpackConfig())
    await fs.remove(compiler.options.output.path)
    return new Promise((resolve, reject) => {
      compiler.run((err, stats) => {
        if (err) return reject(err)
        resolve(stats)
      })
    })
  }

  applyPlugins() {
    const pluginsFromPackage = [
      ...Object.keys(this.projectPkg.data.dependencies || {}),
      ...Object.keys(this.projectPkg.data.devDependencies || {})
    ].filter(name => name.startsWith('poi-plugin-'))

    let plugins = [
      require('./plugins/build'),
      require('./plugins/dev'),
      require('./plugins/config-base'),
      require('./plugins/config-dev'),
      require('./plugins/config-build'),
      require('./plugins/config-app'),
      require('./plugins/config-css'),
      require('./plugins/serve-public'),
      require('poi-plugin-generator')(),
      ...pluginsFromPackage.map(plugin => {
        const pluginName = plugin.replace(/^poi-plugin-/, '')
        return resolveFrom(this.resolveBaseDir(), plugin)(
          this.config.pluginOptions[pluginName] || {}
        )
      })
    ]

    if (this.config.plugins) {
      plugins = plugins.concat(
        parseJsonConfig(this.config.plugins, {
          prefix: 'poi-plugin-'
        })
      )
    }

    this.plugins = plugins

    for (const plugin of plugins) {
      plugin.apply(this)
    }
  }

  hasPlugin(name) {
    return this.plugins && this.plugins.find(plugin => plugin.name === name)
  }

  removePlugin(name) {
    this.plugins = this.plugins.filter(plugin => plugin.name !== name)
    return this
  }

  async run() {
    this.applyPlugins()
    const { input, flags } = await this.cli.parse(this.cliOptions.args)
    if (!this.cli.matchedCommand && !flags.help) {
      if (input[0]) {
        logger.error(
          'Unknown command, run `poi --help` to get a list of available commands.'
        )
      } else {
        this.cli.showHelp()
      }
    }
  }
}

module.exports = opts => new Poi(opts)
