const path = require('path')
const resolveFrom = require('resolve-from')

module.exports = {
  name: 'builtin:config-base',
  apply(api) {
    api.chainWebpack(config => {
      const { pages } = api.config

      for (const entryName of Object.keys(pages)) {
        const { entry } = pages[entryName]
        if (typeof entry === 'string') {
          config.entry(entryName).add(api.resolveBaseDir(entry))
        } else if (Array.isArray(entry)) {
          config.entry(entryName).merge(entry.map(v => api.resolveBaseDir(v)))
        } else {
          throw new TypeError('Entry value must be either string or array!')
        }
      }

      config.merge({
        performance: {
          hints: false
        },
        mode: api.options.command === 'build' ? 'production' : 'development'
      })

      config.output
        .path(api.resolveBaseDir(api.config.outDir))
        // Default values, will later be change by dev and build plugins
        .filename('assets/js/[name].js')
        .publicPath(api.config.publicPath)

      config.resolve.alias.set('@', api.resolveBaseDir('src'))

      // Vue support
      const { VueLoaderPlugin } = require('vue-loader')
      VueLoaderPlugin.__expression = `require('vue-loader').VueLoaderPlugin`
      config.plugin('vue').use(VueLoaderPlugin)

      const templateCompilerPath = resolveFrom.silent(
        api.options.baseDir,
        'vue-template-compiler'
      )
      config.module
        .rule('vue')
        .test(/\.vue$/)
        .use('vue-loader')
        .loader('vue-loader')
        .options({
          compiler: templateCompilerPath && require(templateCompilerPath)
        })

      // Babel support
      config.module
        .rule('js')
        .test(/\.jsx?$/)
        .include.add(filepath => {
          // Don't include node_modules
          return !/node_modules/.test(filepath)
        })
        .end()
        .use('babel-loader')
        .loader(require.resolve('../webpack/loaders/babel-loader'))
        .options({
          // Options for our custom babel-loader
          // Not used for now
          customLoaderOptions: {},
          root: api.resolveBaseDir()
        })

      if (api.cliOptions.progress !== false && !api.cliOptions.debug) {
        const { ProgressPlugin } = require('webpack')
        ProgressPlugin.__expression = `require('webpack').ProgressPlugin`
        config.plugin('progress').use(ProgressPlugin)
      }

      config
        .plugin('report-status')
        .use(require('../webpack/plugins/report-status-plugin'), [
          {
            showFileStats: api.options.command === 'build',
            command: api.options.command,
            devServer: api.config.devServer
          }
        ])

      config.plugin('constants').use(require('webpack').DefinePlugin, [
        Object.assign(
          {
            POI_PUBLIC_PATH: JSON.stringify(api.config.publicPath),
            POI_COMMAND: JSON.stringify(api.command),
            __DEV__: JSON.stringify(api.command !== 'build')
          },
          api.config.constants
        )
      ])

      // Resolve loaders and modules in poi's node_modules folder
      const inWorkspaces = __dirname.includes('/poi/packages/poi/')
      const ownModules = inWorkspaces
        ? path.join(__dirname, '../../../../node_modules')
        : path.join(__dirname, '../../node_modules')
      config.resolve.modules.add(ownModules).add('node_modules')
      config.resolveLoader.modules.add(ownModules).add('node_modules')
    })
  }
}
